/* emu.h -- written by Alexis WILKE (c) 2005-2006 */

/**********************************************************************
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2, or (at your option)
  any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; see the file COPYING.  If not, write to
  the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
**********************************************************************/

#ifndef __EMU65C02_H__
#define __EMU65C02_H__

#include	"emu65xx.h"

#include	<pthread.h>
#include	<sys/time.h>

namespace emulator_8bits
{



class emu65C02 : public emu65xx
{
public:
				emu65C02(processor_t processor, Callback& callback);
				~emu65C02();

	virtual unsigned int	AddRef(void) const;
	virtual unsigned int	Release(void) const;

	virtual void		SetTrace(bool trace = true);
	virtual unsigned int	GetMhz(void) const;
	virtual void		SetMhz(unsigned int mhz);
	virtual speed_t		GetSpeed(void) const;
	virtual void		SetSpeed(speed_t speed);
	virtual void		GetStats(cpu_stats_t& st, bool reset = true) const;

	virtual void		NMI(void);
	virtual void		Reset(void);
	virtual void		Interrupt(void);

// this is a private function which needs to be called from the thread
// startup function...
	void			Run(void);

private:
	// Addressing
	typedef bool (emu65C02::*fetchaddr_t)(void);
	enum addr_t {
		ADDR_NONE,
		ADDR_IMMEDIATE,
		ADDR_PZ,
		ADDR_PZX,
		ADDR_PZY,
		ADDR_INDPZ,
		ADDR_INDPZX,
		ADDR_INDPZY,
		ADDR_ABS,
		ADDR_ABSX,
		ADDR_ABSY,
		ADDR_INDABS,
		ADDR_INDABSX,
		ADDR_RELATIVE,
		ADDR_MAX
	};

	// Instruction definition
	typedef void (emu65C02::*instruction_t)(void);
	struct inst_t {
		unsigned char		f_code;	    		// this is for debug purposes
		const char *		f_name;	    		// name of this instruction in ASCII
		instruction_t		f_inst;	    		// specific instruction
		unsigned char		f_addr : 4; 		// address cycle
		unsigned char		f_load : 1; 		// load cycle
		unsigned char		f_store : 1;		// store cycle
		// if bit 7 set in f_cycles, add 1 if page crossed on address
		// if bit 6 set in f_cycles, add 1 if it branches
		// if bit 5 set in f_cycles, add 1 if decimal mode
		unsigned char		f_cycles[CPU_max];	// cycles (if 0, not available on that CPU)
	};

	// Status register
	struct status_t {
		status_t&		operator = (unsigned char v)
					{
						* (unsigned char *) this = v | 0x20;
					}

		unsigned char		AsChar(void) const
					{
						return * (unsigned char *) this;
					}

		unsigned char		f_carry : 1;		// C
		unsigned char		f_zero : 1;		// Z
		unsigned char		f_interrupt : 1;	// I - 1 to disable
		unsigned char		f_decimal : 1;		// D - 1 for decimal mode
		unsigned char		f_brk : 1;		// B - 0 IRQ, 1 BRK
		unsigned char		f_one : 1;		// always 1
		unsigned char		f_overflow : 1;		// V
		unsigned char		f_negative : 1;		// N
	};

	// the user registers
	struct registers_t {
		unsigned short	f_pc;
		unsigned char	f_a;
		unsigned char	f_x;
		unsigned char	f_y;
		unsigned char	f_s;
		status_t	f_p;
	};

	// To trace our processor
	struct trace_t {
		registers_t	f_regs;		// includes start pc
		unsigned short	f_end;		// after operator & operand
		unsigned short	f_addr;		// address used to load or save or both
		unsigned int	f_cycles;	// the number of cycles for this instruction
	};


	// Address definition
	bool			FetchAddress_NONE(void);
	bool			FetchAddress_IMMEDIATE(void);
	bool			FetchAddress_PZ(void);
	bool			FetchAddress_PZX(void);
	bool			FetchAddress_PZY(void);
	bool			FetchAddress_INDPZ(void);
	bool			FetchAddress_INDPZX(void);
	bool			FetchAddress_INDPZY(void);
	bool			FetchAddress_ABS(void);
	bool			FetchAddress_ABSX(void);
	bool			FetchAddress_ABSY(void);
	bool			FetchAddress_INDABS(void);
	bool			FetchAddress_INDABS6502(void);
	bool			FetchAddress_INDABSX(void);
	bool			FetchAddress_RELATIVE(void);

	// Instructions
	void			inst_ADC(void);
	void			inst_AND(void);
	void			inst_ASL(void);
	void			inst_ASLA(void);
	void			inst_BBC(void);
	void			inst_BBR(void);
	void			inst_BBS(void);
	void			inst_BCC(void);
	void			inst_BCS(void);
	void			inst_BEQ(void);
	void			inst_BIT(void);
	void			inst_BMI(void);
	void			inst_BNE(void);
	void			inst_BPL(void);
	void			inst_BRA(void);
	void			inst_BRK(void);
	void			inst_BVC(void);
	void			inst_BVS(void);
	void			inst_CLC(void);
	void			inst_CLD(void);
	void			inst_CLI(void);
	void			inst_CLV(void);
	void			inst_CMP(void);
	void			inst_CPX(void);
	void			inst_CPY(void);
	void			inst_DEC(void);
	void			inst_DECA(void);
	void			inst_DEX(void);
	void			inst_DEY(void);
	void			inst_EOR(void);
	void			inst_INC(void);
	void			inst_INCA(void);
	void			inst_INX(void);
	void			inst_INY(void);
	void			inst_JMP(void);
	void			inst_JMPIND(void);
	void			inst_JSR(void);
	void			inst_LDA(void);
	void			inst_LDX(void);
	void			inst_LDY(void);
	void			inst_LSR(void);
	void			inst_LSRA(void);
	void			inst_ORA(void);
	void			inst_NOP(void);
	void			inst_PHA(void);
	void			inst_PHP(void);
	void			inst_PHX(void);
	void			inst_PHY(void);
	void			inst_PLA(void);
	void			inst_PLP(void);
	void			inst_PLX(void);
	void			inst_PLY(void);
	void			inst_RMB(void);
	void			inst_ROL(void);
	void			inst_ROLA(void);
	void			inst_ROR(void);
	void			inst_RORA(void);
	void			inst_RTI(void);
	void			inst_RTS(void);
	void			inst_SBC(void);
	void			inst_SEC(void);
	void			inst_SED(void);
	void			inst_SEI(void);
	void			inst_SMB(void);
	void			inst_STA(void);
	void			inst_STP(void);
	void			inst_STX(void);
	void			inst_STY(void);
	void			inst_STZ(void);
	void			inst_TAX(void);
	void			inst_TAY(void);
	void			inst_TRB(void);
	void			inst_TSB(void);
	void			inst_TSX(void);
	void			inst_TXA(void);
	void			inst_TXS(void);
	void			inst_TYA(void);
	void			inst_WAI(void);

	// Utility functions
		// Memory
	void			ResetMap(void);
	unsigned char		Load(unsigned short addr);
	unsigned short		LoadWord(unsigned short addr);
	unsigned short		LoadWordRound(unsigned short addr);
	void			Store(unsigned short addr, unsigned char v);

		// Stack
	void			Push(unsigned char v);
	void			PushWord(unsigned short w);
	unsigned char		Pop(void);
	unsigned short		PopWord(void);

		// Other
	void			SetLoadFlags(unsigned char v);
	void			Trace(const inst_t *inst, trace_t& trace);

	// Reference counter
	mutable unsigned int	f_refcount;

	// The following is defined as the complete list of instructions
	// It is changed depending on the processor in use
	static const inst_t	g_inst[256];

	// the functions to read the address
	fetchaddr_t		f_fetchaddr[ADDR_MAX];

	// Processor
	processor_t		f_processor;

	// CPU registers (public)
	registers_t		f_regs;

	// CPU registers (private)
	unsigned char		f_current_instruction;
	unsigned char		f_temp;
	int			f_addr;
	bool			f_stop;
	bool			f_wait;
	bool			f_trace;
	mutable cpu_stats_t	f_stats;
	int			f_mhz;
	struct timeval		f_start_time;
	speed_t			f_speed;
	long long		f_cycles;		// # of cycles executed (used to sync. the processor properly)
	unsigned int		f_waited_cycles;	// # of cycles we already waited on

	// the callback object
	Callback&		f_callback;

	// the emulator thread & messages
	pthread_t		f_thread;
	pthread_cond_t		f_cond;
	pthread_mutex_t		f_mutex;
	bool			f_quit;
	bool			f_nmi;
	bool			f_reset;
	bool			f_interrupt;
};



};		// emulator_8bits
#endif		// __EMU65C02_H__
// vim: ts=8
